from typing import Optional

from bomf.loader.entityloader import EntityLoader, EntityLoadingResult


class _ExampleEntity:
    pass


class TestEntityLoader:
    class _ExampleEntityLoader(EntityLoader):
        def __init__(self):
            self.sanitize_called: bool = False
            self.verification_called: bool = False
            self.loading_called: bool = False
            self.polling_called: bool = False

        def sanitize(self, entity: _ExampleEntity) -> None:
            assert entity is not None
            self.sanitize_called = True

        async def verify(self, entity: _ExampleEntity, id_in_target_system: Optional[str] = None) -> bool:
            self.verification_called = True
            return True

        async def load_entity(self, entity: _ExampleEntity) -> Optional[EntityLoadingResult]:
            self.loading_called = True
            return EntityLoadingResult(id_in_target_system="foo", polling_task=self.polling_callback("foo"))

        async def polling_callback(self, entity_id: str):
            assert entity_id == "foo"
            self.polling_called = True

    async def test_all_overrides_are_called(self):
        example_loader = TestEntityLoader._ExampleEntityLoader()
        result = await example_loader.load(_ExampleEntity())
        assert example_loader.sanitize_called is True
        assert example_loader.loading_called is True
        assert example_loader.polling_called is True
        assert example_loader.verification_called is True

        assert result.was_loaded_successfully is True
        assert result.loaded_at is not None
        assert result.verified_at is not None
        assert result.verified_at >= result.loaded_at
        assert result.loading_error is None

    async def test_all_overrides_are_called_batch(self):
        example_loader = TestEntityLoader._ExampleEntityLoader()
        result = await example_loader.load_entities([_ExampleEntity()])
        assert example_loader.sanitize_called is True
        assert example_loader.loading_called is True
        assert example_loader.polling_called is True
        assert example_loader.verification_called is True

        assert result[0].was_loaded_successfully is True
        assert result[0].loaded_at is not None
        assert result[0].verified_at is not None
        assert result[0].verified_at >= result[0].loaded_at
        assert result[0].loading_error is None

    async def test_there_is_a_default_sanitize_step(self):
        class _ExampleEntityLoaderWithOutSanitize(EntityLoader):
            # no def sanitize()
            async def verify(self, entity: _ExampleEntity, id_in_target_system: Optional[str] = None) -> bool:
                return True

            async def load_entity(self, entity: _ExampleEntity) -> Optional[EntityLoadingResult]:
                return None

        example_loader = _ExampleEntityLoaderWithOutSanitize()
        result = await example_loader.load(_ExampleEntity())  # must not crash

        assert result.was_loaded_successfully is True
        assert result.loaded_at is not None
        assert result.verified_at is not None
        assert result.verified_at >= result.loaded_at
        assert result.loading_error is None

    async def test_all_overrides_are_called_on_error(self):
        class _ExampleEntityLoaderThatCrashesOnLoad(EntityLoader):
            def __init__(self):
                self.sanitize_called: bool = False
                self.loading_called: bool = False

            def sanitize(self, entity: _ExampleEntity) -> None:
                assert entity is not None
                self.sanitize_called = True

            async def verify(self, entity: _ExampleEntity, id_in_target_system: Optional[str] = None) -> bool:
                raise NotImplementedError()

            async def load_entity(self, entity: _ExampleEntity) -> Optional[EntityLoadingResult]:
                self.loading_called = True
                raise ValueError("Something is wrong")

        example_loader = _ExampleEntityLoaderThatCrashesOnLoad()
        result = await example_loader.load(_ExampleEntity())
        assert example_loader.sanitize_called is True
        assert example_loader.loading_called is True

        assert result.was_loaded_successfully is False
        assert result.loaded_at is None
        assert result.verified_at is None
        assert isinstance(result.loading_error, ValueError) is True



"""
entity loaders load entities into the target system
"""
import asyncio
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Awaitable, Generic, List, Optional, TypeVar

import attrs

_TargetEntity = TypeVar("_TargetEntity")


@attrs.define(auto_attribs=True, kw_only=True)
class EntityLoadingResult:  # pylint:disable=too-few-public-methods
    """
    Information gathered while loading a _TargetEntity into the target system.
    """

    id_in_target_system: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)), default=None
    )
    """
    the optional ID of the entity in the target system (e.g. if a new (GU)ID is generated upon loading)
    """
    polling_task: Optional[Awaitable] = attrs.field(default=None)
    """
    If this task is awaited it means, that the target system is done with processing the request.
    A possible use case is that the target system responds with something like an event ID which can be used to poll
    an endpoint until it returns the expected result.
    """


@attrs.define(auto_attribs=True, kw_only=True)
class LoadingSummary(ABC, Generic[_TargetEntity]):  # pylint:disable=too-few-public-methods
    """
    Each instance of _TargetEntity that is loaded to the target system results in a LoadingSummary.
    It is a summary that reports to calling code.
    """

    was_loaded_successfully: bool = attrs.field(validator=attrs.validators.instance_of(bool))
    """
    true iff the instance has been loaded successfully
    """
    loaded_at: Optional[datetime] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(datetime)), default=None
    )
    """
    point in time at which the loading (without verification) has completed; if not None
    """
    verified_at: Optional[datetime] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(datetime)), default=None
    )
    """
    point in time at which the loading of this entity has been verified (or None if not)
    """
    id_in_target_system: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)), default=None
    )
    """
    the optional ID of the entity in the target system (e.g. if a new (GU)ID is generated upon loading)
    """
    loading_error: Optional[Exception] = attrs.field(default=None)


class EntityLoader(ABC, Generic[_TargetEntity]):  # pylint:disable=too-few-public-methods
    """
    An instance of an entity loader loads data into a target system.
    There is one loader class per target system entity.
    """

    @abstractmethod
    async def load_entity(self, entity: _TargetEntity) -> Optional[EntityLoadingResult]:
        """
        Load the given entity into the target system.
        This method shall contain the code that accesses the target system.
        If the method raises no exception the loading is interpreted to be successful.
        The method may return an EntityLoadingResult if it can provide the information contained therein or None.
        """

    @abstractmethod
    async def verify(self, entity: _TargetEntity, id_in_target_system: Optional[str] = None) -> bool:
        """
        Verify that the given entity has been successfully loaded into the target system.
        Returns true iff the target system knows the given entity (meaning: the loading was successful).
        """

    # pylint:disable=unused-argument
    def sanitize(self, entity: _TargetEntity) -> None:
        """
        sanitize the given entity, by default (no override) does nothing
        """
        return

    async def load(self, entity: _TargetEntity) -> LoadingSummary:
        """
        Loads the given entity into the target system and verifies it has been loaded.
        """

        self.sanitize(entity)
        entity_loading_result: EntityLoadingResult
        loaded_at: datetime
        try:
            entity_loading_result_or_none = await self.load_entity(entity)
            if entity_loading_result_or_none is None:
                # default to a entity loading result that contains no information
                entity_loading_result = EntityLoadingResult(id_in_target_system=None, polling_task=None)
            else:
                entity_loading_result = entity_loading_result_or_none
            loaded_at = datetime.utcnow()
        except Exception as loading_error:  # pylint:disable=broad-except
            # this block is intended to be a Pokemon catcher
            return LoadingSummary(
                id_in_target_system=None,
                was_loaded_successfully=False,
                loading_error=loading_error,
                loaded_at=None,
            )
        if entity_loading_result.polling_task is not None:
            await entity_loading_result.polling_task
        verification_result = await self.verify(
            entity=entity, id_in_target_system=entity_loading_result.id_in_target_system
        )
        return LoadingSummary(
            was_loaded_successfully=verification_result,
            verified_at=datetime.utcnow(),
            loaded_at=loaded_at,
            loading_error=None,
        )

    async def load_entities(self, entities: List[_TargetEntity]) -> List[LoadingSummary]:
        """
        load all the given entities into the target system
        """
        # here we could use some error handling in the future
        tasks: List[Awaitable[LoadingSummary]] = [self.load(entity) for entity in entities]
        result = await asyncio.gather(*tasks)
        return result



from pathlib import Path
from typing import List

import pytest  # type:ignore[import]

from bomf.provider import JsonFileSourceDataProvider, KeyTyp, ListBasedSourceDataProvider, SourceDataProvider


class LegacyDataSystemDataProvider(SourceDataProvider):
    """
    a dummy for access to a legacy system from which we want to migrate data
    """

    def get_entry(self, key: KeyTyp) -> str:
        raise NotImplementedError("Not relevant for this test")

    def get_data(self) -> List[str]:
        return ["foo", "bar", "baz"]


class TestSourceDataProvider:
    def test_provider(self):
        # this is a pretty dumb test
        provider_under_test = LegacyDataSystemDataProvider()
        assert isinstance(provider_under_test.get_data(), list)

    @pytest.mark.datafiles("./unittests/example_source_data.json")
    def test_json_file_provider(self, datafiles):
        file_path = datafiles / Path("example_source_data.json")
        example_json_data_provider = JsonFileSourceDataProvider(
            file_path,
            data_selector=lambda d: d["data"],  # type:ignore[call-overload]
            key_selector=lambda d: d["myKey"],  # type:ignore[index]
        )
        assert example_json_data_provider.get_data() == [
            {"myKey": "hello", "asd": "fgh"},
            {"myKey": "world", "qwe": "rtz"},
        ]
        assert example_json_data_provider.get_entry("world") == {"myKey": "world", "qwe": "rtz"}
        with pytest.raises(KeyError):
            _ = example_json_data_provider.get_entry("something unknown")


class TestListBasedSourceDataProvider:
    def test_list_based_provider(self):
        my_provider = ListBasedSourceDataProvider(["foo", "bar", "baz"], key_selector=lambda x: x)
        assert len(my_provider.get_data()) == 3
        assert my_provider.get_entry("bar") == "bar"



"""
general data models for migrations
"""
import enum
import uuid
from abc import ABC, abstractmethod
from typing import Iterable, Optional, Type, TypeVar, Union

import attrs
from bo4e.bo.geschaeftsobjekt import Geschaeftsobjekt
from bo4e.com.com import COM

_SpecificBusinessObject = TypeVar("_SpecificBusinessObject", bound=Geschaeftsobjekt)
"""
an arbitrary but fixed business object type
"""

_SpecificCom = TypeVar("_SpecificCom", bound=COM)
"""
an arbitrary but fixed COM type
"""

Bo4eTyp = Union[_SpecificBusinessObject, _SpecificCom]


# pylint:disable=too-few-public-methods
@attrs.define(kw_only=True, auto_attribs=True)
class BusinessObjectRelation:
    """
    A business object relation describes the relation between two business object.
    E.g. a relation could have the type "has_melo" where relation_part_a is a bo4e.bo.Vertrag
    and relation_part_b is a bo4e.bo.Messlokation. Some relations are already defined in BO4E itself (e.g MaLo/MeLo)
    or MeLo/Address.
    The idea is to not enforce too much of a structure to the downstream code but still push coders to think about
    necessary relation information.
    """

    relation_type: enum.Enum = attrs.field()
    """
    The relation type describes how two business objects relate to each other.
    This is not (only) about cardinality. It's about being able to model different relations between objects.
    Think about e.g. a business partner and an address: The relation could be:
    - the address is the residential address of the business partner
    - the address is the invoice address of the business partner
    - the address is the place where the business partner was born
    All these relation types are 1:1 relations between business partners and adresses, yet they all carry different
    meaning which we'd like to distinguish in our data.
    """
    relation_part_a: Bo4eTyp = attrs.field()
    """
    one Business Object or COM
    """

    relation_part_b: Bo4eTyp = attrs.field()
    """
    another Business Object or COM
    """


class Bo4eDataSet(ABC):
    """
    A BO4E data set is a collection of Business Objects that relate to each other.
    """

    def __init__(self):
        self._uuid = uuid.uuid4()

    def get_id(self) -> str:
        """
        returns a unique id that only this dataset uses.
        Inheriting classes may overwrite this behaviour, if the dataset has "more natural" keys.
        By default, the id is a stringified UUID that is created in the dataset constructor.
        """
        try:
            return str(self._uuid)
        except AttributeError as attribute_error:
            if attribute_error.name == "_uuid" in str(attribute_error):
                raise ValueError(
                    f"You probably forgot to call super().__init()__ in the constructor of {self.__class__}"
                ) from attribute_error
            raise



"""
filters can be used to consider only those objects for a migration that meet certain conditions
"""

# pylint:disable=too-few-public-methods

import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Awaitable, Generic, List, TypeVar

Candidate = TypeVar("Candidate")  #: an arbitrary but fixed type on which the filter operates


class Filter(ABC, Generic[Candidate]):
    """
    A filter takes objects from candidates and returns only those that are relevant for its use case.
    The advantage over inlining the filtering is that we can easily test these and log in the well-defined filters.
    """

    def __init__(self):
        self._logger = logging.getLogger(self.__module__)

    @abstractmethod
    async def predicate(self, candidate: Candidate) -> bool:
        """
        Returns true iff the candidate shall pass the filter.

        You might wonder why this method is async: The plan is that this allows for building filters that do not only
        depend on the data that you want to filter itself but on other things as well. The most obvious "other thing"
        is the data situation in the target system to which we migrate.
        If we migrate from system A to system B, then one possible filter is to check "does system B know my record"?
        These kind of questions are generally answered in an async way.
        """
        raise NotImplementedError("The inheriting class has to implement this method")

    async def apply(self, candidates: List[Candidate]) -> List[Candidate]:
        """
        apply this filter on the candidates
        """
        tasks: List[Awaitable[bool]] = [self.predicate(c) for c in candidates]
        self._logger.info("%s created %i predicate tasks; Awaiting them all", str(self), len(tasks))
        predicate_results = await asyncio.gather(*tasks)
        self._logger.info("%s awaited %i tasks", str(self), len(tasks))
        result = [
            c for c, predicate_match in zip(candidates, predicate_results, strict=True) if predicate_match is True
        ]
        candidates_removed = sum(1 for pr in predicate_results if pr is False)
        self._logger.info(
            "%i out of %i candidates have been removed by the filter", candidates_removed, len(candidates)
        )
        return result


Aggregate = TypeVar("Aggregate")
"""
Aggregate is a type that enhances the candidate with some additional information used by a Filter on the aggregate.
It has to be possible to extract the candidate from the aggregate.
"""


class AggregateFilter(ABC, Generic[Candidate, Aggregate]):
    """
    A filter that takes objects from candidates and returns only those that are relevant for its use case.
    The difference to the plain Filter is, that the conditions can be defined on groups of candidates instead of
    isolated single candidates. Still it (other than e.g. a usual groupby+filter) logs those entries that do not pass
    the filter.
    """

    def __init__(self, base_filter: Filter[Aggregate]):
        """
        Instantiate by providing a filter that is applied on the aggregate
        """
        self._logger = logging.getLogger(self.__module__)
        self._base_filter = base_filter

    @abstractmethod
    async def aggregate(self, candidates: List[Candidate]) -> List[Aggregate]:
        """
        Create aggregates which are then passed to the base filter that works on the aggregate.
        The method is async so that you can do complex (and e.g. network based) aggregations.
        """
        raise NotImplementedError("The inheriting class has to implement this method")

    @abstractmethod
    def disaggregate(self, aggregate: Aggregate) -> Candidate:
        """
        extract a single candidate from the aggregate that passed the filter
        """
        raise NotImplementedError("The inheriting class has to implement this method")

    async def apply(self, candidates: List[Candidate]) -> List[Candidate]:
        """
        If aggregate and disaggregate and the base_filter are properly setup, then apply will filter the list of
        candidates based on the aggregate base filter.
        """
        aggregates = await self.aggregate(candidates)
        self._logger.info("There are %i candidates and %i aggregates", len(candidates), len(aggregates))
        filtered_aggregates = await self._base_filter.apply(aggregates)
        self._logger.info("There are %i filtered aggregates left", len(filtered_aggregates))
        return [self.disaggregate(fa) for fa in filtered_aggregates]



import enum
from typing import Iterable, Optional, Type

import pytest  # type:ignore[import]
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.com.adresse import Adresse

from bomf.model import Bo4eDataSet, Bo4eTyp, BusinessObjectRelation


class _GeschaeftspartnerAdresseRelation(enum.Enum):
    HAS_LIEFERANSCHRIFT = 1
    HAS_RECHNUNGSANSCHRIFT = 2
    HAS_GEBURTSORT = 3


class _ExampleDataSet(Bo4eDataSet):
    def __init__(self):
        super().__init__()
        self.business_partner = Geschaeftspartner.construct(name1="Müller", name2="Klaus")
        self.address = Adresse.construct(strasse="Rechnungsstrasse", hausnummer="5")

    def get_relations(self) -> Iterable[BusinessObjectRelation]:
        return [
            BusinessObjectRelation(
                relation_type=_GeschaeftspartnerAdresseRelation.HAS_LIEFERANSCHRIFT,
                relation_part_a=self.business_partner,
                relation_part_b=self.address,
            )
        ]

    def get_business_object(self, bo_type: Type[Bo4eTyp], specification: Optional[str] = None) -> Bo4eTyp:
        # pyling:disable=fixme
        # todo: find out how to allow the static type checker to not complain about the "dynamic" type
        if bo_type == Geschaeftspartner:
            return self.business_partner  # type:ignore[return-value]
        if bo_type == Adresse:
            return self.address  # type:ignore[return-value]
        raise NotImplementedError(f"The bo type {bo_type} is not implemented")


class TestBo4eDataSet:
    async def test_example_data_set(self):
        dataset: _ExampleDataSet = _ExampleDataSet()
        assert len(list(dataset.get_relations())) == 1
        assert isinstance(dataset.get_business_object(Geschaeftspartner), Geschaeftspartner)
        assert isinstance(dataset.get_business_object(Adresse), Adresse)
        assert dataset.get_id() is not None



"""
mappers convert from source data model to BO4E and from BO4E to a target data model
"""
from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar

from bomf.model import Bo4eDataSet

TargetDataModel = TypeVar("TargetDataModel")
"""
Target data model is the data model of the target (meaning: the data model of the system to which you'd like to migrate)
"""

IntermediateDataSet = TypeVar("IntermediateDataSet", bound=Bo4eDataSet)
"""
Intermediate data set is the BO4E based layer between source and target.
It is based on BO4E.
"""


# pylint:disable=too-few-public-methods
class SourceToBo4eDataSetMapper(ABC, Generic[IntermediateDataSet]):
    """
    A mapper that maps one or multiple sources into Bo4eDataSets
    """

    # the inheriting class is free to combine and bundle the source data as it wants.
    # the only thing it has to provide is a method to create_data_sets (in bo4e).
    # we don't care from where it gets them in the first place

    def create_data_sets(self) -> List[IntermediateDataSet]:
        """
        apply the mapping to all the provided source data sets.

        """
        raise NotImplementedError("The inheriting class has to implement this method")


# pylint:disable=too-few-public-methods
class Bo4eDataSetToTargetMapper(ABC, Generic[TargetDataModel, IntermediateDataSet]):
    """
    A mapper that transforms data from the intermediate bo4e model to the target data model
    """

    @abstractmethod
    def create_target_model(self, dataset: IntermediateDataSet) -> TargetDataModel:
        """
        maps the given source data model into an intermediate data set
        """

    def create_target_models(self, datasets: List[IntermediateDataSet]) -> List[TargetDataModel]:
        """
        apply the mapping to all the provided dataset
        """
        # here we could use some error handling in the future
        return [self.create_target_model(dataset=dataset) for dataset in datasets]



"""
providers provide data
"""
import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, Generic, List, Mapping, Optional, Protocol, TypeVar, Union

SourceDataModel = TypeVar("SourceDataModel")
"""
Source data model is the data model of the source (meaning: the data model of the system from which the data originate).
"""

KeyTyp = TypeVar("KeyTyp")
"""
The type of the key used as "primary key" for the source data model
"""


# pylint:disable=too-few-public-methods
class SourceDataProvider(ABC, Generic[SourceDataModel, KeyTyp]):
    """
    A source data provider provides entities from the source data system.
    The source data provider is thought to encapsulate data access behind a unified interface.
    """

    @abstractmethod
    def get_data(self) -> List[SourceDataModel]:
        """
        Returns all available entities from the source data model.
        They will be filtered in a SourceDataModel Filter ("Preselect")
        """

    @abstractmethod
    def get_entry(self, key: KeyTyp) -> SourceDataModel:
        """
        returns the source data model which has key as key.
        raises an error if the key is unknown
        """


class ListBasedSourceDataProvider(SourceDataProvider[SourceDataModel, KeyTyp]):
    """
    A source data provider that is instantiated with a list of source data models
    """

    def __init__(self, source_data_models: List[SourceDataModel], key_selector: Callable[[SourceDataModel], KeyTyp]):
        """
        instantiate it by providing a list of source data models
        """
        self._models: List[SourceDataModel] = source_data_models
        self._models_dict: Mapping[KeyTyp, SourceDataModel] = {key_selector(m): m for m in source_data_models}
        self.key_selector = key_selector

    def get_entry(self, key: KeyTyp) -> SourceDataModel:
        return self._models_dict[key]

    def get_data(self) -> List[SourceDataModel]:
        return self._models


class JsonFileSourceDataProvider(SourceDataProvider[SourceDataModel, KeyTyp], Generic[SourceDataModel, KeyTyp]):
    """
    a source data model provider that is based on a JSON file
    """

    def __init__(
        self,
        json_file_path: Path,
        data_selector: Callable[[Union[dict, list]], List[SourceDataModel]],
        key_selector: Callable[[SourceDataModel], KeyTyp],
        encoding="utf-8",
    ):
        """
        initialize by providing a filepath to the json file and an accessor that describes the position of the data
        within the file.
        """
        with open(json_file_path, "r", encoding=encoding) as json_file:
            raw_data = json.load(json_file)
        self._source_data_models: List[SourceDataModel] = data_selector(raw_data)
        self._key_to_data_model_mapping: Mapping[KeyTyp, SourceDataModel] = {
            key_selector(sdm): sdm for sdm in self._source_data_models
        }
        self.key_selector = key_selector

    def get_data(self) -> List[SourceDataModel]:
        return self._source_data_models

    def get_entry(self, key: KeyTyp) -> SourceDataModel:
        return self._key_to_data_model_mapping[key]



"""
BOMF stands for BO4E Migration Framework.
"""
from abc import ABC
from typing import Generic, List

import attrs

from bomf.filter import Filter
from bomf.loader.entityloader import EntityLoader, LoadingSummary
from bomf.mapper import Bo4eDataSetToTargetMapper, IntermediateDataSet, SourceToBo4eDataSetMapper, TargetDataModel
from bomf.provider import KeyTyp, SourceDataProvider
from bomf.validation import Bo4eDataSetValidation


# pylint:disable=too-few-public-methods
@attrs.define(kw_only=True, auto_attribs=True)
class MigrationStrategy(ABC, Generic[IntermediateDataSet, TargetDataModel]):
    """
    A migration strategy describes the whole migration flow of datasets from a source to a target system
    """

    source_data_set_to_bo4e_mapper: SourceToBo4eDataSetMapper[IntermediateDataSet]
    """
    A mapper that transforms source data models into data sets that consist of bo4e objects
    """
    validation: Bo4eDataSetValidation[IntermediateDataSet]
    """
    a set of validation rules that are applied to the bo4e data sets
    """
    bo4e_to_target_mapper: Bo4eDataSetToTargetMapper[TargetDataModel, IntermediateDataSet]
    """
    a mapper that transforms bo4e data sets to a structure that suits the target system
    """
    target_loader: EntityLoader[TargetDataModel]
    """
    The target loader moves the target entities into the actual target system.
    """

    async def migrate(self) -> List[LoadingSummary]:
        """
        run the entire migration flow from source to target which includes:
        1. create bo4e data source using the source_data_set_to_bo4e_mapper
        2. checking that all the bo4e data sets obey the validation rules
        3. mapping from bo4e to the target data model
        4. loading the target data models into the target system.
        """
        # todo: here we should add some logging and statistics stuff
        bo4e_datasets = self.source_data_set_to_bo4e_mapper.create_data_sets()
        valid_entries = self.validation.validate(bo4e_datasets).valid_entries
        target_data_models = self.bo4e_to_target_mapper.create_target_models(valid_entries)
        loading_summaries = await self.target_loader.load_entities(target_data_models)
        return loading_summaries



"""
Source Data Provider Filters combine the features of a filter with the features of a Source Data Provider.
"""

from typing import Callable, Generic, List, Literal, Optional, overload

from bomf import KeyTyp, SourceDataProvider
from bomf.filter import Candidate, Filter
from bomf.provider import JsonFileSourceDataProvider, ListBasedSourceDataProvider


# pylint:disable=too-few-public-methods
class SourceDataProviderFilter(Generic[Candidate, KeyTyp]):
    """
    a filter that works on and returns a CandidateSourceDataProvider
    """

    def __init__(self, candidate_filter: Filter[Candidate]):
        """
        instantiate by providing a filter which can be applied on the data providers source data models
        """
        self._filter = candidate_filter

    @overload
    async def apply(self, source_data_provider: JsonFileSourceDataProvider) -> SourceDataProvider[Candidate, KeyTyp]:
        ...

    @overload
    async def apply(self, source_data_provider: ListBasedSourceDataProvider) -> SourceDataProvider[Candidate, KeyTyp]:
        ...

    @overload
    async def apply(
        self, source_data_provider: JsonFileSourceDataProvider, key_selector: Literal[None]
    ) -> SourceDataProvider[Candidate, KeyTyp]:
        ...

    @overload
    async def apply(
        self, source_data_provider: ListBasedSourceDataProvider, key_selector: Literal[None]
    ) -> SourceDataProvider[Candidate, KeyTyp]:
        ...

    async def apply(
        self,
        source_data_provider: SourceDataProvider[Candidate, KeyTyp],
        key_selector: Optional[Callable[[Candidate], KeyTyp]] = None,
    ) -> SourceDataProvider[Candidate, KeyTyp]:
        """
        Reads all the data from the given source_data_provider, applies the filtering, then returns a new source
        data provider that only contains those entries that passed the filter (its predicate).

        If the provided source_data_provider is a JsonFileSourceDataProvider, then you don't have to provide a
        key_selector (let it default to None).
        However, in general, you have to specify how the data can be indexed using a key_selector which is not None.
        If you provide both a JsonFileSourceDataProvider AND a key_selector, the explicit key_selector will be used.
        """
        survivors: List[Candidate] = await self._filter.apply(source_data_provider.get_data())
        key_selector_to_be_used: Callable[[Candidate], KeyTyp]
        if key_selector is not None:
            key_selector_to_be_used = key_selector
        else:
            key_selector_to_be_used = source_data_provider.key_selector  # type:ignore[attr-defined]
            # if this raises an attribute error you have to
            # * either provide a source_data_provider which has a key_selector attribute
            # * or explicitly provide a key_selector as (non-None) argument
        filtered_data_provider_class = ListBasedSourceDataProvider(
            source_data_models=survivors, key_selector=key_selector_to_be_used
        )
        return filtered_data_provider_class



import logging
from itertools import cycle
from typing import Iterable, Optional, Type

import pytest  # type:ignore[import]
from bo4e.bo.geschaeftspartner import Geschaeftspartner
from bo4e.bo.marktlokation import Marktlokation
from bo4e.enum.verbrauchsart import Verbrauchsart

from bomf.model import Bo4eDataSet, Bo4eTyp, BusinessObjectRelation
from bomf.validation import Bo4eDataSetRule, Bo4eDataSetValidation, DataSetTyp, DataSetValidationResult


class GeschaeftspartnerMesslokatinDataSet(Bo4eDataSet):
    """
    a dummy data set which contains a Geschaeftspartner and a Messlokation
    """

    def __init__(self, geschaeftspartner: Geschaeftspartner, marktlokation: Marktlokation):
        super().__init__()
        self._gp = geschaeftspartner
        self._malo = marktlokation

    def get_business_object(self, bo_type: Type[Bo4eTyp], specification: Optional[str] = None) -> Bo4eTyp:
        if bo_type == Geschaeftspartner:
            return self._gp
        if bo_type == Marktlokation:
            return self._malo
        raise NotImplementedError("Not relevant for this test")

    def get_relations(self) -> Iterable[BusinessObjectRelation]:
        raise NotImplementedError("Not relevant for this test")

    def get_id(self) -> str:
        return self._malo.marktlokations_id


class DontMigrateCustomersWithVornameKlausAndWaermenutzung(Bo4eDataSetRule):
    """
    This test should show that the validations act on a whole data set that consists of multiple objects.
    The validations are thought to catch inconsitencies between the objects while the (source data) filters should
    be responsible for single objects only.
    """

    def validate(self, dataset: GeschaeftspartnerMesslokatinDataSet) -> DataSetValidationResult:
        customer: Geschaeftspartner = dataset.get_business_object(Geschaeftspartner)
        malo: Marktlokation = dataset.get_business_object(Marktlokation)
        forbidden_verbrauchsarts_for_Klaus = (Verbrauchsart.W, Verbrauchsart.WS, Verbrauchsart.KLW, Verbrauchsart.KLWS)
        if customer.name2 == "Klaus" and malo.verbrauchsart in forbidden_verbrauchsarts_for_Klaus:
            return DataSetValidationResult(is_valid=False, error_message="hier die fehlermeldung")
        return DataSetValidationResult(is_valid=True)

    def __str__(self):
        return "Kein Klaus mit Wärme"


class ARuleThatCrashesInOneOutOfFourTimesForDemoPurposes(Bo4eDataSetRule):
    def __init__(self):
        self._should_crash = cycle([False, False, True, False])

    def validate(self, dataset: DataSetTyp) -> DataSetValidationResult:
        should_crash = next(self._should_crash)
        if should_crash:
            raise Exception("something went terribly wrong")
        return DataSetValidationResult(is_valid=True)

    def __str__(self):
        return "schlecht programmiert"


class MyValidation(Bo4eDataSetValidation):
    """
    all the validation rules I'd like the datasets to obey
    """

    def __init__(self):
        super().__init__(
            rules=[
                DontMigrateCustomersWithVornameKlausAndWaermenutzung(),
                ARuleThatCrashesInOneOutOfFourTimesForDemoPurposes(),
            ]
        )


class TestValidation:
    def test_validation(self, caplog):
        candidates = [
            GeschaeftspartnerMesslokatinDataSet(
                # valid, because verbrauchsart is no Wärme (W)
                marktlokation=Marktlokation.construct(marktlokations_id="53502368955", verbrauchsart=Verbrauchsart.KL),
                geschaeftspartner=Geschaeftspartner.construct(name2="Klaus"),
            ),
            GeschaeftspartnerMesslokatinDataSet(
                # invalid, because verbrauchsart is also Wärme (W)
                marktlokation=Marktlokation.construct(marktlokations_id="87301147632", verbrauchsart=Verbrauchsart.KLW),
                geschaeftspartner=Geschaeftspartner.construct(name2="Klaus"),
            ),
            GeschaeftspartnerMesslokatinDataSet(
                # valid, because vorname is not Klaus but crashes in the second rule
                marktlokation=Marktlokation.construct(marktlokations_id="78192756766", verbrauchsart=Verbrauchsart.W),
                geschaeftspartner=Geschaeftspartner.construct(name2="Günther"),
            ),
            GeschaeftspartnerMesslokatinDataSet(
                # valid, because vorname is not Klaus - no crash
                marktlokation=Marktlokation.construct(marktlokations_id="18410127695", verbrauchsart=Verbrauchsart.W),
                geschaeftspartner=Geschaeftspartner.construct(name2="Klause"),
            ),
        ]
        caplog.set_level(logging.DEBUG, logger=self.__module__)
        my_validation = MyValidation()
        validation_result = my_validation.validate(candidates)
        assert len(validation_result.valid_entries) == 2
        assert len(validation_result.invalid_entries) == 2
        assert "dataset 53502368955 obeys the rule 'Kein Klaus mit Wärme'" in caplog.messages
        assert "✔ data set 53502368955 is valid" in caplog.messages
        assert "dataset 87301147632 does not obey: 'hier die fehlermeldung'" in caplog.messages
        assert "❌ data set 78192756766 is invalid" in caplog.messages
        assert "Validation of rule 'schlecht programmiert' on dataset 78192756766 failed" in caplog.messages
        assert "❌ data set 78192756766 is invalid" in caplog.messages



"""
bomf performs validation on the intermediate bo4e data set layer.
"""
import logging
from abc import ABC, abstractmethod
from typing import Generic, Iterable, List, Optional, Tuple, TypeVar

import attrs

from bomf.model import Bo4eDataSet

# pylint:disable=too-few-public-methods
DataSetTyp = TypeVar("DataSetTyp", bound=Bo4eDataSet)


@attrs.define(kw_only=True, auto_attribs=True)
class DataSetValidationResult:
    """
    a dataset validation result is the outcome of the evaluation of one atomic validation rule
    """

    is_valid: bool = attrs.field(validator=attrs.validators.instance_of(bool))
    """
    true iff the data set passed the validation
    """
    error_message: Optional[str] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(str)), default=None
    )
    """
    The error message should be not None and contain a descriptive error message if is_valid is False.
    """


class Bo4eDataSetRule(ABC, Generic[DataSetTyp]):
    """
    A bo4e dataset rule performs exactly one atomic check on a dataset.
    It is thought to ensure the consistency within a data set (in contrast to the self-consistency of the single objects
    inside the data set which should be part of the filters on the source data models (aka "pre-select")).
    """

    @abstractmethod
    def validate(self, dataset: DataSetTyp) -> DataSetValidationResult:
        """
        checks if the given dataset is valid
        """
        raise NotImplementedError("The inheriting class shall implement this method")

    # inheriting classes should overwrite __str__ for pretty log messages


@attrs.define(kw_only=True, auto_attribs=True)
class _ValidAndInvalidEntities(Generic[DataSetTyp]):
    """
    A container type that holds both the invalid and the valid entries.
    """

    # This class is only used internally and shouldn't be imported elsewhere. That's why its clunky name doesn't matter.

    valid_entries: List[DataSetTyp] = attrs.field(default=attrs.Factory(list))
    """
    those entries that are valid and may pass on to the loader
    """
    invalid_entries: List[Tuple[DataSetTyp, List[str]]] = attrs.field(default=attrs.Factory(list))
    """
    those entries that are invalid together with their respective error messages
    """


@attrs.define(kw_only=True, auto_attribs=True)
class Bo4eDataSetValidation(ABC, Generic[DataSetTyp]):
    """
    A Bo4e Dataset Validation consists of multiple rules that are checked one after another.
    """

    rules: List[Bo4eDataSetRule[DataSetTyp]] = attrs.field()
    """
    the rules which a single data set should obey
    """

    def validate(self, datasets: Iterable[DataSetTyp]) -> _ValidAndInvalidEntities:
        """
        applies all rules to all datasets
        """
        _logger = logging.getLogger(self.__module__)
        result: _ValidAndInvalidEntities[DataSetTyp] = _ValidAndInvalidEntities()
        for dataset in datasets:
            dataset_id = dataset.get_id()
            error_messages: List[str] = []
            for rule in self.rules:
                validation_result: DataSetValidationResult
                try:
                    validation_result = rule.validate(dataset)
                except Exception:  # pylint:disable=broad-except # pokemon catcher is intended
                    error_message = f"Validation of rule '{rule}' on dataset {dataset_id} failed"
                    _logger.exception(error_message, exc_info=True)
                    error_messages.append(error_message)
                    continue
                if validation_result.is_valid:
                    _logger.debug("dataset %s obeys the rule '%s'", dataset_id, str(rule))
                else:
                    _logger.debug("dataset %s does not obey: '%s'", dataset_id, validation_result.error_message)
                    error_messages.append(validation_result.error_message or "<no error message provided>")
            if len(error_messages) == 0:
                result.valid_entries.append(dataset)
                _logger.info("✔ data set %s is valid", dataset_id)
            else:
                result.invalid_entries.append((dataset, error_messages))
                _logger.info("❌ data set %s is invalid", dataset_id)

        return result



import dataclasses
import logging
from itertools import groupby
from typing import List

import pytest  # type:ignore[import]

from bomf.filter import AggregateFilter, Filter
from bomf.filter.sourcedataproviderfilter import SourceDataProviderFilter
from bomf.provider import ListBasedSourceDataProvider, SourceDataProvider


class _FooFilter(Filter):
    async def predicate(self, candidate: dict) -> bool:
        return "foo" in candidate and candidate["foo"] == "bar"


class TestFilter:
    @pytest.mark.parametrize(
        "filter_under_test,candidates,survivors",
        [
            pytest.param(
                _FooFilter(),
                [{"foo": "baz"}, {"foo": "bar"}],
                [{"foo": "bar"}],
            ),
        ],
    )
    async def test_filter(self, filter_under_test: Filter, candidates: List[dict], survivors: List[dict], caplog):
        caplog.set_level(logging.DEBUG, logger=self.__module__)
        actual = await filter_under_test.apply(candidates)
        assert actual == survivors
        assert "1 out of 2 candidates have been removed by the filter" in caplog.messages


@dataclasses.dataclass
class _MyCandidate:
    number: int
    string: str


@dataclasses.dataclass
class _MyAggregate:
    group_key: str
    max_number_for_key: int
    candidate: _MyCandidate


class _BarFilter(AggregateFilter):
    """
    An Aggregate Filter that groups _MyCandidates by their string attribute and keeps only those entries that have the
    highest number (attribute) in their respective group.
    It's basically a show-case test that allows to understand how the aggregate filters are supposed to be used.
    """

    def __init__(self):
        class _BaseFilter(Filter[_MyAggregate]):
            async def predicate(self, candidate: _MyAggregate) -> bool:
                return candidate.max_number_for_key == candidate.candidate.number

        base_filter = _BaseFilter()
        super(_BarFilter, self).__init__(base_filter)

    async def aggregate(self, candidates: List[_MyCandidate]) -> List[_MyAggregate]:
        result: List[_MyAggregate] = []
        for group_key, group in groupby(sorted(candidates, key=lambda c: c.string), lambda c: c.string):
            group_items = list(group)
            max_number_in_group = max(group_item.number for group_item in group_items)
            for group_item in group_items:
                result.append(
                    _MyAggregate(group_key=group_key, max_number_for_key=max_number_in_group, candidate=group_item)
                )
        return result

    def disaggregate(self, aggregate: _MyAggregate) -> _MyCandidate:
        return aggregate.candidate


class TestAggregateFilter:
    @pytest.mark.parametrize(
        "filter_under_test,candidates,survivors",
        [
            pytest.param(
                _BarFilter(),
                [
                    _MyCandidate(number=1, string="foo"),
                    _MyCandidate(number=19, string="bar"),
                    _MyCandidate(number=2, string="foo"),
                    _MyCandidate(number=17, string="bar"),
                ],
                [_MyCandidate(number=19, string="bar"), _MyCandidate(number=2, string="foo")],
            ),
        ],
    )
    async def test_aggregate_filter(
        self, filter_under_test: AggregateFilter, candidates: List[dict], survivors: List[dict], caplog
    ):
        caplog.set_level(logging.DEBUG, logger=self.__module__)
        actual = await filter_under_test.apply(candidates)
        assert actual == survivors
        assert "There are 4 candidates and 4 aggregates" in caplog.messages
        assert "There are 2 filtered aggregates left" in caplog.messages


class TestSourceDataProviderFilter:
    @pytest.mark.parametrize(
        "candidate_filter,candidates,survivors",
        [
            pytest.param(
                _BarFilter(),
                [
                    _MyCandidate(number=1, string="foo"),
                    _MyCandidate(number=19, string="bar"),
                    _MyCandidate(number=2, string="foo"),
                    _MyCandidate(number=17, string="bar"),
                ],
                [_MyCandidate(number=19, string="bar"), _MyCandidate(number=2, string="foo")],
            ),
        ],
    )
    async def test_source_data_provider_filter(
        self,
        candidate_filter: Filter[_MyCandidate],
        candidates: List[_MyCandidate],
        survivors: List[_MyCandidate],
        caplog,
    ):
        my_provider: ListBasedSourceDataProvider[_MyCandidate, int] = ListBasedSourceDataProvider(
            candidates, key_selector=lambda mc: mc.number
        )
        sdp_filter: SourceDataProviderFilter[_MyCandidate, int] = SourceDataProviderFilter(candidate_filter)
        caplog.set_level(logging.DEBUG, logger=self.__module__)
        filtered_provider = await sdp_filter.apply(my_provider)
        assert isinstance(filtered_provider, SourceDataProvider)
        actual = filtered_provider.get_data()
        assert actual == survivors
        assert "There are 4 candidates and 4 aggregates" in caplog.messages
        assert "There are 2 filtered aggregates left" in caplog.messages

    async def test_source_data_provider_filter_error(self):
        my_provider = ListBasedSourceDataProvider([{"foo": "bar"}, {"foo": "notbar"}], key_selector=lambda d: d["foo"])
        del my_provider.key_selector
        sdp_filter: SourceDataProviderFilter[_MyCandidate, int] = SourceDataProviderFilter(_FooFilter())
        with pytest.raises(AttributeError):
            await sdp_filter.apply(my_provider)



"""
Tests the overall data flow using bomf.
"""
from typing import Dict, List, Optional

import attrs

from bomf import (
    Bo4eDataSetToTargetMapper,
    Bo4eDataSetValidation,
    EntityLoader,
    Filter,
    MigrationStrategy,
    SourceDataProvider,
    SourceToBo4eDataSetMapper,
)
from bomf.loader.entityloader import EntityLoadingResult
from bomf.model import Bo4eDataSet
from bomf.provider import KeyTyp, SourceDataModel
from bomf.validation import Bo4eDataSetRule, DataSetValidationResult

_MySourceDataModel = Dict[str, str]
_MyKeyTyp = str
_MyTargetDataModel = List[str]


@attrs.define(auto_attribs=True, kw_only=True)
class _MyIntermediateDataModel(Bo4eDataSet):
    data: Dict[str, str]

    def get_id(self) -> str:
        return "12345"


class _MySourceDataProvider(SourceDataProvider[_MySourceDataModel, _MyKeyTyp]):
    def get_entry(self, key: KeyTyp) -> _MySourceDataModel:
        raise NotImplementedError("Not relevant for the test")

    def get_data(self) -> List[_MySourceDataModel]:
        return [
            {"foo": "bar"},
            {"FOO": "BAR"},
            {"Foo": "Bar"},
            {"remove by filter": "should not pass the filter"},
            {"invalid": "doesn't matter"},
        ]


class _MyFilter(Filter[_MySourceDataModel]):
    async def predicate(self, candidate: _MySourceDataModel) -> bool:
        return "remove by filter" not in candidate


class _MyToBo4eMapper(SourceToBo4eDataSetMapper[_MyIntermediateDataModel]):
    def __init__(self, what_ever_you_like: List[_MySourceDataModel]):
        # what_ever_you_like is a place holde for all the relation magic that may happen
        self._source_models = what_ever_you_like

    def create_data_sets(self) -> List[_MyIntermediateDataModel]:
        return [_MyIntermediateDataModel(data=source) for source in self._source_models]


class _MyRule(Bo4eDataSetRule[_MyIntermediateDataModel]):
    def validate(self, dataset: _MyIntermediateDataModel) -> DataSetValidationResult:
        return DataSetValidationResult(is_valid="invalid" not in dataset.data)


class _MyValidation(Bo4eDataSetValidation):
    def __init__(self):
        super().__init__(rules=[_MyRule()])


class _MyToTargetMapper(Bo4eDataSetToTargetMapper[_MyTargetDataModel, _MyIntermediateDataModel]):
    def create_target_model(self, dataset: _MyIntermediateDataModel) -> _MyTargetDataModel:
        my_dict = dataset.data
        for my_key, my_value in my_dict.items():
            return [my_key, my_value]
        return ["doesnt", "matter"]


class _MyTargetLoader(EntityLoader):
    async def load_entity(self, entity: _MyTargetDataModel) -> Optional[EntityLoadingResult]:
        async def polling():
            return True

        return EntityLoadingResult(id_in_target_system="Fooooo", polling_task=polling())

    async def verify(self, entity: _MyTargetDataModel, id_in_target_system: Optional[str] = None) -> bool:
        return True


class MyMigrationStrategy(MigrationStrategy):
    pass


class TestMigrationStrategy:
    """
    This is more of an integration than a unit test. All the single components come together here.
    """

    async def test_happy_path(self):
        # here's some pre-processing, you can read some data, you can create relations, whatever
        raw_data = _MySourceDataProvider().get_data()
        survivors = await _MyFilter().apply(raw_data)
        to_bo4e_mapper = _MyToBo4eMapper(what_ever_you_like=survivors)
        strategy = MyMigrationStrategy(
            source_data_set_to_bo4e_mapper=to_bo4e_mapper,
            validation=_MyValidation(),
            bo4e_to_target_mapper=_MyToTargetMapper(),
            target_loader=_MyTargetLoader(),
        )
        result = await strategy.migrate()
        assert result is not None
        assert len(result) == 3  # = source models -1(filter) -1(validation)



from typing import List, Optional, Type

import attrs
import pytest  # type:ignore[import]
from bo4e.bo.marktlokation import Marktlokation
from bo4e.bo.messlokation import Messlokation

from bomf.mapper import Bo4eDataSetToTargetMapper, SourceToBo4eDataSetMapper
from bomf.model import Bo4eTyp


class _NotImplementedBo4eDataSetMixin:
    """
    a mixin to inherit from if you'd like to have correct types but don't care about the logic
    """


@attrs.define(kw_only=True, auto_attribs=True)
class _MaLoAndMeLo(_NotImplementedBo4eDataSetMixin):
    malo: Marktlokation = attrs.field()
    melo: Messlokation = attrs.field()

    def get_business_object(self, bo_type: Type[Bo4eTyp], specification: Optional[str] = None) -> Bo4eTyp:
        # pyling:disable=fixme
        # todo: find out how to allow the static type checker to not complain about the "dynamic" type
        if bo_type == Marktlokation:
            return self.malo  # type:ignore[return-value]
        if bo_type == Messlokation:
            return self.melo  # type:ignore[return-value]
        raise NotImplementedError(f"The bo type {bo_type} is not implemented")


# in these tests we assume, that:
# - the source data model is a dictionary
# - the intermediate data model are BO4E MaLo and MeLo
# - the target data model is a list of string
# This is just to demonstrate the mapping structures.


class _DictToMaLoMeLoMapper(SourceToBo4eDataSetMapper):
    def create_data_sets(self) -> List[_MaLoAndMeLo]:
        return [
            _MaLoAndMeLo(
                melo=Messlokation.construct(messlokations_id=source["meloId"]),
                malo=Marktlokation.construct(marktlokations_id=source["maloId"]),
            )
            for source in [{"maloId": "54321012345", "meloId": "DE000111222333"}]
        ]


class _MaLoMeLoToListMapper(Bo4eDataSetToTargetMapper):
    def create_target_model(self, dataset: _MaLoAndMeLo) -> List[str]:
        return [
            dataset.get_business_object(Marktlokation).marktlokations_id,
            dataset.get_business_object(Messlokation).messlokations_id,
        ]


class TestMapper:
    def test_source_to_intermediate_mapper_batch(self):
        mapper = _DictToMaLoMeLoMapper()
        actual = mapper.create_data_sets()
        assert actual == [
            _MaLoAndMeLo(
                melo=Messlokation.construct(messlokations_id="DE000111222333"),
                malo=Marktlokation.construct(marktlokations_id="54321012345"),
            )
        ]

    def test_intermediate_to_target_mapper(self):
        """
        tests the single data set mapping
        """
        mapper = _MaLoMeLoToListMapper()
        actual = mapper.create_target_model(
            _MaLoAndMeLo(
                melo=Messlokation.construct(messlokations_id="DE000111222333"),
                malo=Marktlokation.construct(marktlokations_id="54321012345"),
            )
        )
        assert actual == ["54321012345", "DE000111222333"]

    def test_intermediate_to_target_mapper_batch(self):
        """
        test the batch mapping
        """
        mapper = _MaLoMeLoToListMapper()
        actual = mapper.create_target_models(
            [
                _MaLoAndMeLo(
                    melo=Messlokation.construct(messlokations_id="DE000111222333"),
                    malo=Marktlokation.construct(marktlokations_id="54321012345"),
                )
            ]
        )
        assert actual == [["54321012345", "DE000111222333"]]
